/*
 * uvector - uniform (homogeneous) numeric vector types (SRFI-4)
 *
 *  (c) 1999 Shiro Kawai  shiro@acm.org
 *  Terms and conditions to use, modify or distribute this code
 *  is the same as of STk.
 *
 *  $Id: uvector.c,v 1.2 1999/11/08 11:22:24 shiro Exp $
 */

#include "uvector.h"

extern void s8vector_init();
extern void u8vector_init();
extern void s16vector_init();
extern void u16vector_init();
extern void s32vector_init();
extern void u32vector_init();
extern void s64vector_init();
extern void u64vector_init();
extern void f32vector_init();
extern void f64vector_init();

static void uvector_mark(SCM);
static void uvector_free(SCM);
static void uvector_display(SCM, SCM, int);
static SCM uvector_compare(SCM, SCM, int);

extern uvector_subtype s8vector_desc;
extern uvector_subtype u8vector_desc;
extern uvector_subtype s16vector_desc;
extern uvector_subtype u16vector_desc;
extern uvector_subtype s32vector_desc;
extern uvector_subtype u32vector_desc;
extern uvector_subtype s64vector_desc;
extern uvector_subtype u64vector_desc;
extern uvector_subtype f32vector_desc;
extern uvector_subtype f64vector_desc;

int tc_uvector;

static uvector_subtype *uvector_subtypes[] = 
{
    &s8vector_desc,
    &u8vector_desc,
    &s16vector_desc,
    &u16vector_desc,
    &s32vector_desc,
    &u32vector_desc,
    &s64vector_desc,
    &u64vector_desc,
    &f32vector_desc,
    &f64vector_desc
};

static STk_extended_scheme_type uvector_typedesc = {
    "uniform-vector",           /* name */
    0,                          /* procp */
    uvector_mark,               /* gc_mark_fct */
    uvector_free,               /* gc_sweep_fct */
    NULL,                       /* apply_fct */
    uvector_display,            /* display_fct */
    uvector_compare             /* compare_fct */
};

static void uvector_mark(SCM obj)
{
    void (*fn)(SCM) = uvector_subtypes[UVECTOR_TYPE(obj)]->mark;
    if (fn) fn(obj);
}

static void uvector_free(SCM obj)
{
    free(EXTDATA(obj));
}

static void uvector_display(SCM obj, SCM port, int mode)
{
    void (*fn)(SCM, SCM, int) = uvector_subtypes[UVECTOR_TYPE(obj)]->display;
    if (fn) fn(obj, port, mode);
}

static SCM uvector_compare(SCM a, SCM b, int equalp)
{
    SCM (*fn)(SCM, SCM, int) = uvector_subtypes[UVECTOR_TYPE(a)]->compare;
    if (fn) return fn(a, b, equalp);
    else return Truth;
}

static int uvector_tag_to_type(SCM stag, char *proc_name)
{
    int i;
    char *tag = NULL;
    
    if (STRINGP(stag)) {
        tag = CHARS(stag);
    } else if (SYMBOLP(stag)) {
        tag = PNAME(stag);
    } else {
        Serror("symbol or string is expected", stag);
    }
    for (i=0; i<UVECTOR_TYPE_MAX; i++) {
        if (strcmp(uvector_subtypes[i]->tagname, tag) == 0) {
            return i;
        }
    }
    Serror("invalid subtype for uniform vector", stag);
    return -1;                  /* NOTREACHED */
}

SCM STk_uvector_p(SCM obj)
{
    return UVECTORP(obj)? Truth : Ntruth;
}

SCM STk_make_uvector(SCM tag, SCM length, SCM init)
{
    static char *proc_name = "make-uvector";
    int subtype = uvector_tag_to_type(tag, proc_name);
    return uvector_subtypes[subtype]->make_from_length(length, init);
}

SCM STk_uvector(SCM args, int len)
{
    static char *proc_name = "uvector";
    SCM tag;
    int subtype;
    if (len < 1) Serror("a tag is required", NIL);
    tag = CAR(args);
    subtype = uvector_tag_to_type(tag, proc_name);
    return uvector_subtypes[subtype]->make_from_args(CDR(args), len - 1);
}

void STk_init_uvector()
{
    tc_uvector = STk_add_new_type(&uvector_typedesc);

    s8vector_init();
    u8vector_init();
    s16vector_init();
    u16vector_init();
    s32vector_init();
    u32vector_init();
    s64vector_init();
    u64vector_init();
    f32vector_init();
    f64vector_init();

    STk_add_new_primitive("uvector?", tc_subr_1,
                          (PRIMITIVE(*)())STk_uvector_p);
    STk_add_new_primitive("make-uvector", tc_subr_2_or_3,
                          (PRIMITIVE(*)())STk_make_uvector);
    STk_add_new_primitive("uvector", tc_lsubr,
                          (PRIMITIVE(*)())STk_uvector);
}
